#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>

#include "types.hpp"
#include "types_isd.hpp"
#include "manager.hpp"
#include "tools.hpp"
#include "iSDOS.hpp"
#include "iSDOS_tools.hpp"
#include "lang.hpp"

extern PluginStartupInfo startupInfo;
extern Options           op;

Manager::Manager(const wchar_t* fileName, FmtPlugin* fmtPlugin)
{
  lstrcpyW(hostFileName, fileName);
  filer = fmtPlugin;
  img = filer->openSubPlugin(fileName);
  *curDir    = 0;
  ZeroMemory(&lastModifed, sizeof(lastModifed));

  filer->reload(img);
  openHostFile();
  //    
  u8 buf[blockSize];
  readBlock(0, buf);
  CopyMemory(&dsk, buf, sizeof(dsk));

  //   
  readFolder(dsk.mainDir1stBlock, files);

  for(int fNum = 1; fNum < files[0].dir.totalFiles; ++fNum)
  {
    if(files[fNum].attr & FLAG_EXIST && !isDir(files[fNum]))
    {
      if(strncmp(files[fNum].name, "device  sys", 11) == 0)
      {
        CopyMemory(&device_sys_hdr, &files[fNum], sizeof(UniHdr));
        u32 size = getSize(device_sys_hdr);
        
        int noBlocks = size/blockSize + ((size%blockSize) ? 1 : 0);
        device_sys = (u8*)malloc(blockSize*noBlocks);
        read_device_sys();
        break;
      }
    }
  }

  closeHostFile();
}

Manager::~Manager()
{
  free(device_sys);
  filer->closeSubPlugin(img);
}

void Manager::GetOpenPanelInfo(struct OpenPanelInfo *info)
{
  info->StructSize = sizeof(*info);
  info->Flags      = /*OPIF_USEFILTER | OPIF_USEHIGHLIGHTING |*/ OPIF_ADDDOTS | OPIF_SHOWNAMESONLY | OPIF_SHOWPRESERVECASE;
  info->HostFile   = hostFileName;
  info->CurDir     = curDir;
  info->Format     = L"iS-DOS";

  static wchar_t panelTitle[300];

  *panelTitle = 0;
  lstrcatW(panelTitle, L" iS-DOS:");
  lstrcatW(panelTitle, pointToName(hostFileName));
  lstrcatW(panelTitle, curDir);
  lstrcatW(panelTitle, L" ");
  info->PanelTitle = panelTitle;

  static PanelMode mode[5];

  static wchar_t columnTitle0[] = L"               ";
  static wchar_t columnTitle1[] = L"               ";
  static wchar_t columnTitle2[] = L"               ";
  static wchar_t columnTitle3[] = L"               ";
  static wchar_t columnTitle4[] = L"               ";
  static wchar_t *columnTitles[] = { columnTitle0, columnTitle1, columnTitle2, columnTitle3, columnTitle4, };

  const wchar_t* p = op.columnTitles;
  int n = 0;
  int i = 0;
  while(*p)
  {
    if(*p != L',')
      columnTitles[n][i++] = *p++;
    else
    {
      ++p;
      columnTitles[n++][i] = 0;
      i = 0;
      if(n == 5) break;
    }
  }
  
  mode[4].ColumnTypes        = op.columnTypes;
  mode[4].ColumnWidths       = op.columnWidths;
  mode[4].ColumnTitles       = columnTitles;
  mode[4].StatusColumnTypes  = L"NR,SC,DMBM";
  mode[4].StatusColumnWidths = L"0,8,0";
  mode[4].Flags              = PMFLAGS_ALIGNEXTENSIONS;

  info->PanelModesArray  = mode;
  info->PanelModesNumber = sizeof(mode)/sizeof(mode[0]);
  info->StartPanelMode   = '0' + op.defaultPanelMode;
  if(readInfo())
  {
    static InfoPanelLine infoLines[4]{};
    int noChars;
    if(dsk.signatureChic[0] == 'D' && dsk.signatureChic[1] == 'S' && dsk.signatureChic[2] == 'K')
    {
      noChars = 11;
      infoLines[0].Text = GetMsg(MChic);
    }
    else
    {
      noChars = 8;
      infoLines[0].Text = GetMsg(MClassic);
    }
    infoLines[0].Flags = IPLFLAGS_SEPARATOR;
    
    infoLines[1].Text = GetMsg(MDiskTitle);
    while(noChars--)
      if(dsk.title[noChars] != ' ' && dsk.title[noChars] != 0) break;

    static wchar_t Data1[100];
    swprintf(Data1, L"%.*hs", noChars + 2, dsk.title);
    infoLines[1].Data = Data1;
    infoLines[1].Flags = IPLFLAGS_NONE;

    static wchar_t Data2[100];
    infoLines[2].Text = GetMsg(MTotalBlocks);
    wsprintfW(Data2, L"%d", dsk.noBlocks);
    infoLines[2].Data = Data2;
    infoLines[2].Flags = IPLFLAGS_NONE;

    static wchar_t Data3[100];
    infoLines[3].Text = GetMsg(MFreeBlocks);
    wsprintfW(Data3, L"%d", noFreeBlocks);
    infoLines[3].Data = Data3;
    infoLines[3].Flags = IPLFLAGS_NONE;
    info->InfoLines = infoLines;
    info->InfoLinesNumber = sizeof(infoLines)/sizeof(infoLines[0]);
  }

}

int Manager::getFindData(PluginPanelItem **pPanelItem, size_t *pNoItems, int opMode)
{
  *pNoItems   = 0;
  *pPanelItem = NULL;

  if(!readInfo()) return FALSE;

  PluginPanelItem* item = new PluginPanelItem[files[0].dir.noFiles]{};

  int inx = 0;

  for(int fNum = 1; fNum < files[0].dir.totalFiles; ++fNum)
  {
    if(files[fNum].attr & FLAG_EXIST)
    {
      item[inx].CustomColumnNumber = 3;
      wchar_t ** CustomColumnData = new LPWSTR[item[0].CustomColumnNumber];
      item[inx].CustomColumnData = CustomColumnData;
      
      CustomColumnData[0] = new wchar_t[5];
      CustomColumnData[1] = new wchar_t[6];
      CustomColumnData[2] = new wchar_t[6];

      FILETIME ft{};
      if(isDir(files[fNum]))
      {
        DosDateTimeToFileTime(files[fNum].date, 0, &ft);
        item[inx].FileAttributes = FILE_ATTRIBUTE_DIRECTORY;
        if(files[fNum].attr & FLAG_SOLID)
          wsprintfW(CustomColumnData[1], L"%5d", files[fNum].dir.firstBlock);
        else
          wsprintfW(CustomColumnData[1], L"%5d", files[fNum].dir.descr1stBlock);
        CustomColumnData[2][0] = 0;
      }
      else
      {
        item[inx].FileAttributes = FILE_ATTRIBUTE_NORMAL;
        item[inx].FileSize     = getSize(files[fNum]);
        wsprintfW(CustomColumnData[1], L"%5d", files[fNum].file.firstBlock);
        wsprintfW(CustomColumnData[2], L"%5d", files[fNum].file.loadAddr);
        DosDateTimeToFileTime(files[fNum].date, files[fNum].file.time, &ft);
      }
      if(files[fNum].attr & FLAG_HIDDEN)
        item[inx].FileAttributes |= FILE_ATTRIBUTE_HIDDEN;

      LocalFileTimeToFileTime(&ft, &ft);

      makeName(files[fNum], pcFileName[fNum]);

      item[inx].FileName = pcFileName[fNum];

      item[inx].LastWriteTime = ft;

      lstrcpyW(CustomColumnData[0], L"....");
      if(!(files[fNum].attr & FLAG_SOLID))       CustomColumnData[0][0] = L'S';
      if(files[fNum].attr & FLAG_READ_PROTECT)   CustomColumnData[0][1] = L'R';
      if(files[fNum].attr & FLAG_WRITE_PROTECT)  CustomColumnData[0][2] = L'W';
      if(files[fNum].attr & FLAG_DELETE_PROTECT) CustomColumnData[0][3] = L'D';
      ++inx;
    }
  }
  *pNoItems   = inx;
  *pPanelItem = item;
  return TRUE;
}

void Manager::freeFindData(PluginPanelItem *panelItem, int  noItems)
{
    for(int i = 0; i < noItems; i++)
    {
        for(int j = 0; j < panelItem[i].CustomColumnNumber; ++j)
        {
            delete[] panelItem[i].CustomColumnData[j];
        }
        delete[] panelItem[i].CustomColumnData;
    }
    delete[] panelItem;
}

int Manager::setDirectory(const wchar_t* dirName, int opMode)
{
  u8 buf[blockSize];
  if(*dirName == L'\\')
    if(*(++dirName) == 0)
    {
      *curDir = 0;
      if(!openHostFile()) return FALSE;
      readFolder(dsk.mainDir1stBlock, files);
      closeHostFile();
      return TRUE;
    }
  if(!lstrcmpW(dirName, L".."))
  {
    if(*curDir == 0) return FALSE;

    if(!openHostFile()) return FALSE;
    readFolder(files[0].dir.parentDir1stBlock, files);
    closeHostFile();

    int i = lstrlenW(curDir) - 2;
    while(i != -1 && curDir[i] != L'\\') --i;
    curDir[i] = 0;
  }
  else
  {
    int fNum = 1;
    wchar_t name[8+3+2];
    for(;fNum < files[0].dir.totalFiles; ++fNum)
    {
      if(!(files[fNum].attr & FLAG_EXIST)) continue;
      if(!isDir(files[fNum])) continue;
      makeName(files[fNum], name);
      if(!lstrcmpW(name, dirName)) break;
    }
    if(fNum == files[0].dir.totalFiles) return FALSE;

    if(!openHostFile()) return FALSE;
    readFolder(files[fNum], files);
    closeHostFile();

    lstrcatW(curDir, L"\\");
    lstrcatW(curDir, dirName);
  }
  return TRUE;
}
