#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <memory>
#include <string>

#include "plugin.hpp"

#include "types.hpp"
#include "types_scl.hpp"
#include "manager.hpp"
#include "tools.hpp"
#include "lang.hpp"

extern Options op;
extern PluginStartupInfo startupInfo;
extern const GUID MainGuid;

void addEndSlash(std::wstring &path);

int Manager::createFile(HANDLE& file, const wchar_t* name, int& action)
{
  //   
  WIN32_FIND_DATAW data;
  HANDLE h = FindFirstFileW(name, &data);
  if(h != INVALID_HANDLE_VALUE)
  {
    //     
    FindClose(h);
    if(action == 2) return 0;
    if(action == 0)
    {
      const wchar_t *msgItems[11];
      msgItems[0] = GetMsg(MWarning);
      msgItems[1] = GetMsg(MAlreadyExists);
      msgItems[2] = name;
      msgItems[3] = L"\x001";
      msgItems[5] = L"\x001";
      msgItems[6] = GetMsg(MOverwrite);
      msgItems[7] = GetMsg(MAll);
      msgItems[8] = GetMsg(MSkip);
      msgItems[9] = GetMsg(MSkipAll);
      msgItems[10] = GetMsg(MCancel);

      wchar_t param[50];
      if(data.ftLastWriteTime.dwLowDateTime == 0 &&
         data.ftLastWriteTime.dwHighDateTime == 0)
      {
        wsprintfW(param, GetMsg(MDestination1), data.nFileSizeLow);
      }
      else
      {
        FILETIME lastWriteTime;
        FileTimeToLocalFileTime(&data.ftLastWriteTime, &lastWriteTime);
        
        SYSTEMTIME time;
        FileTimeToSystemTime(&lastWriteTime, &time);
        
        wsprintfW(param, GetMsg(MDestination2),
                 data.nFileSizeLow,
                 time.wDay, time.wMonth, time.wYear,
                 time.wHour, time.wMinute, time.wSecond);
      }
      msgItems[4] = param;
      int askCode = messageBox(FMSG_WARNING /*| FMSG_DOWN*/,
                               msgItems,
                               std::size(msgItems), 5);
      switch(askCode)
      {
        case -1:
        case 4:
                  return -1;
        case 3:
                  action = 2; //skipAll
        case 2:
                  return 0;
        case 1:
                  action = 1; //overwriteAll
      }
    }
  }
  file = CreateFileW(name,
                    GENERIC_WRITE,
                    FILE_SHARE_READ,
                    NULL,
                    CREATE_ALWAYS,
                    FILE_FLAG_SEQUENTIAL_SCAN,
                    NULL);
  if(file != INVALID_HANDLE_VALUE) return 1;
  
  const wchar_t *msgItems[4];
  msgItems[0] = GetMsg(MError);
  msgItems[1] = GetMsg(MCanNotCreate);
  msgItems[2] = name;
  msgItems[3] = GetMsg(MOk);
  messageBox(FMSG_WARNING, msgItems, std::size(msgItems), 1);
  return 0;
}

int Manager::getFiles(PluginPanelItem *panelItem,
                      size_t noItems,
                      int move,
                      const wchar_t *destPath,
                      int opMode)
{
  if(noItems == 0) return 1;
  if(isReadOnly && move)
  {
      const wchar_t *msgItems[4];
    msgItems [0] = GetMsg(MWarning);
    msgItems [1] = GetMsg(MCanNotMove);
    msgItems [2] = GetMsg(MWriteProtected);
    msgItems [3] = GetMsg(MOk);
    messageBox(FMSG_WARNING, msgItems, std::size(msgItems), 1);
    return FALSE;
  }
  std::wstring DestPathCopy(destPath);
  if(!DestPathCopy.empty()) addEndSlash(DestPathCopy);

  wchar_t msg[100];
  const wchar_t *action[2];
  const wchar_t *action2[2];
  const wchar_t*button[2];
  action[0] = GetMsg(MCopy);
  action[1] = GetMsg(MMove);
  action2[0] = GetMsg(MCopying);
  action2[1] = GetMsg(MMoving);
  button[0] = GetMsg(MCopyButton);
  button[1] = GetMsg(MMoveButton);
  if(noItems == 1)
    wsprintfW(msg,
             GetMsg(MFileTo),
             move ? action2[1] : action2[0],
             panelItem[0].FileName);
  else
    wsprintfW(msg, GetMsg(MFilesTo), move ? action2[1] : action2[0], noItems);

  const wchar_t historyName[] = L"xSCL_copy_path";
  
  FarDialogItem dialogItems[]={
    { DI_DOUBLEBOX,   3,  1, 72, 11, {FALSE},     nullptr, nullptr,                                   0, move ? action[1] : action[0] },
    { DI_TEXT,        5,  2,  0,  0, {FALSE},     nullptr, nullptr,                                   0, msg},
    { DI_EDIT,        5,  3, 70,  3,  {TRUE}, historyName, nullptr,                         DIF_HISTORY, DestPathCopy.c_str() },
    { DI_TEXT,        3,  4,  0,  0, {FALSE},     nullptr, nullptr,        DIF_BOXCOLOR | DIF_SEPARATOR, L"" },
    { DI_RADIOBUTTON, 5,  5,  0,  0, {FALSE},     nullptr, nullptr,                           DIF_GROUP, GetMsg(MHoBeta) },
    { DI_RADIOBUTTON, 5,  6,  0,  0, {FALSE},     nullptr, nullptr,                                   0, GetMsg(MSCL) },
    { DI_TEXT,        3,  7,  0,  0, {FALSE},     nullptr, nullptr,        DIF_BOXCOLOR | DIF_SEPARATOR, L"" },
    { DI_CHECKBOX,    5,  8,  0,  0, {FALSE},     nullptr, nullptr,                                   0, GetMsg(MSkipHeaders) },
    { DI_TEXT,        3,  9,  0,  0, {FALSE},     nullptr, nullptr,        DIF_BOXCOLOR | DIF_SEPARATOR, L"" },
    { DI_BUTTON,      0, 10,  0,  0, {FALSE},     nullptr, nullptr, DIF_CENTERGROUP | DIF_DEFAULTBUTTON, move ? button[1] : button[0] },
    { DI_BUTTON,      0, 10,  0,  0, {FALSE},     nullptr, nullptr,                     DIF_CENTERGROUP, GetMsg(MCancel) }
  };


  switch(op.defaultFormat)
  {
    case 1:
              dialogItems[4].Selected = TRUE;
              break;
    case 2:
              dialogItems[5].Selected = TRUE;
              break;
    default:
              if(noItems == 1)
                dialogItems[4].Selected = TRUE;
              else
                dialogItems[5].Selected = TRUE;
  }
  
  bool singleFile = false, skipHeader = false;
  //    ,  
  if((opMode & OPM_SILENT) == 0)
  {
      std::unique_ptr<void, decltype(SafeDlgFree) *> Dlg{ startupInfo.DialogInit(&MainGuid,
                                       &__uuidof(struct __declspec(uuid("{2DCEFDFB-2292-488C-ACB8-E30C424AF1FD}")) TDlgGuid),
                                       -1, -1, 76, 13,
                                       NULL,
                                       dialogItems,
                                       std::size(dialogItems), 0, 0, startupInfo.DefDlgProc, nullptr), SafeDlgFree };
      if(Dlg.get() == INVALID_HANDLE_VALUE)
      {
          return -1;
      }
      intptr_t askCode = startupInfo.DialogRun(Dlg.get());

    if(askCode != 9) return -1;
    lstrcpyW(DestPathCopy.data(), dialogItems[2].Data);
    
    singleFile  = !startupInfo.SendDlgMessage(Dlg.get(), DM_GETCHECK, 4, 0);
    skipHeader = startupInfo.SendDlgMessage(Dlg.get(), DM_GETCHECK, 7, 0);
  }
  
  //   ,   
  if(GetFileAttributesW(DestPathCopy.c_str()) == 0xFFFFFFFF)
  {
      for(wchar_t *c = DestPathCopy.data(); *c; c++)
      {
          if(*c != L' ')
          {
              for(; *c; c++)
              {
                  if(*c == L'\\')
                  {
                      *c = 0;
                      CreateDirectoryW(DestPathCopy.c_str(), NULL);
                      *c = L'\\';
                  }
              }
              CreateDirectoryW(DestPathCopy.c_str(), NULL);
              break;
          }
      }
  }
  if(!DestPathCopy.empty() && DestPathCopy.back() != ':') addEndSlash(DestPathCopy);

  DWORD noBytesWritten;
  DWORD checkSum     = 0;
  
  int buttonStatus = 0; //   overwrite_all/skip_all
  
  HANDLE file;
  wchar_t   name[300];
  
  if(singleFile)
  {
    wsprintfW(name, L"%s%s", DestPathCopy.c_str(), panelItem[0].FileName);
    *pointToExt(name) = 0;
    if(skipHeader)
      lstrcatW(name, L"bin");
    else
      lstrcatW(name, L"scl");
    int r = createFile(file, name, buttonStatus);
    if(r == 0 || r == -1) return -1;
    if(!skipHeader) checkSum = writeSCLHeader(file, noItems);
  }
  
  int  exitCode       = 1;
  BYTE noFilesWritten = 0;
  
  if(!openHostFile()) return 0;
  
  HANDLE screen = startupInfo.SaveScreen(0,0,-1,-1);

  for(int iNum = 0; iNum < noItems; ++iNum)
  {
    //   
    int fNum;
    for(fNum = 0; fNum < noFiles; ++fNum)
      if(!lstrcmpW(panelItem[iNum].FileName, pcFiles[fNum].name)) break;

    //     
    if(noItems == 1 && ((opMode & OPM_VIEW) || (opMode & OPM_EDIT)) && pcFiles[fNum].skipHeader) skipHeader = true;

    if(!singleFile)
    {
      wsprintfW(name, L"%s%s", DestPathCopy.c_str(), panelItem[iNum].FileName);
      if(skipHeader) lstrcatW(name, L".bin");
      int r = createFile(file, name, buttonStatus);
      if(r == -1)
      {
        exitCode = -1;
        break;
      }
      if(r == 0)
      {
        exitCode = 0;
        continue;
      }
      if(!skipHeader)
      {
        HoHdr hdr;
        CopyMemory(hdr.name, files[fNum].name, 8);
        
        hdr.type     = files[fNum].type;
        hdr.start    = files[fNum].start;
        hdr.size     = files[fNum].size;
        hdr.reserved = 0;
        hdr.noSecs   = files[fNum].noSecs;
        hdr.checkSum = calculateCheckSum(hdr);

        WriteFile(file, &hdr, sizeof(HoHdr), &noBytesWritten, NULL);
      }
    }
    else
    {
      if(!skipHeader)
      {
        //  
        SetFilePointer(file, iNum*sizeof(SCLHdr)+8+1, NULL, FILE_BEGIN);
        WriteFile     (file, &files[fNum], sizeof(SCLHdr), &noBytesWritten, NULL);
        checkSum += calculateCheckSum((BYTE*)&files[fNum], sizeof(SCLHdr));
      }
    }
    
    const wchar_t* msgItems[2]{ GetMsg(MCopyingFile), name };
    wchar_t  shortName[51] = L"...";
    if(lstrlenW(name) > 47)
    {
      lstrcatW(shortName, name + lstrlenW(name) - 47);
      msgItems[1] = shortName;
    }
    if((opMode & OPM_SILENT) == 0)
      messageBox(0, msgItems, std::size(msgItems), 0);

    //  
    checkSum += copyFile(fNum, file, skipHeader);
    if(!singleFile) CloseHandle(file);
    
    //     
    panelItem[iNum].Flags ^= PPIF_SELECTED;
    if(move) pcFiles[fNum].deleted = true;
    ++noFilesWritten;
  }
  closeHostFile();

  startupInfo.RestoreScreen(screen);

  if(singleFile)
  {
    if(!skipHeader)
      WriteFile(file, &checkSum, sizeof(checkSum), &noBytesWritten, NULL);
    CloseHandle(file);
  }
  if(move) deleteFilesImpl(noFilesWritten);
  
  return exitCode;
}
