#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <memory>
#include <algorithm>

#include "plugin.hpp"

#include "types.hpp"
#include "types_scl.hpp"
#include "manager.hpp"
#include "tools.hpp"
#include "lang.hpp"

extern PluginStartupInfo startupInfo;
extern const GUID MainGuid;

const DWORD PKF_ANY = (LEFT_CTRL_PRESSED | RIGHT_CTRL_PRESSED) | SHIFT_PRESSED | (LEFT_ALT_PRESSED | RIGHT_ALT_PRESSED);

int Manager::processKey(const INPUT_RECORD *Rec)
{
    if(Rec->EventType != KEY_EVENT)
    {
        return FALSE;
    }

    if(!Rec->Event.KeyEvent.bKeyDown)
    {
        return FALSE;
    }
    auto key = Rec->Event.KeyEvent.wVirtualKeyCode;
    auto controlState = Rec->Event.KeyEvent.dwControlKeyState;

  if(((controlState & SHIFT_PRESSED) && key == VK_F6) ||
     (!(controlState & PKF_ANY) && key == VK_F4) ||
     ((controlState & (LEFT_CTRL_PRESSED | RIGHT_CTRL_PRESSED)) && key == 0x41))
  {
    if(isReadOnly)
    {
        const wchar_t *msgItems[4];
      msgItems [0] = GetMsg(MWarning);
      msgItems [1] = GetMsg(MCanNotEdit);
      msgItems [2] = GetMsg(MWriteProtected);
      msgItems [3] = GetMsg(MOk);
      messageBox(FMSG_WARNING, msgItems, std::size(msgItems), 1);
      return TRUE;
    }
    HANDLE screen = startupInfo.SaveScreen(0,0,-1,-1);
    PanelInfo info = { sizeof(PanelInfo) };
    startupInfo.PanelControl(this, FCTL_GETPANELINFO, 0, &info);
    startupInfo.RestoreScreen(screen);

    if(info.ItemsNumber > 0 && info.CurrentItem > 0)
    {
      int fNum = 0;
      intptr_t Size = startupInfo.PanelControl(this, FCTL_GETCURRENTPANELITEM, 0, nullptr);
      std::unique_ptr<PluginPanelItem, decltype(::free) *> Item((PluginPanelItem *)::malloc(Size), ::free);
      FarGetPluginPanelItem GetItem = { sizeof(FarGetPluginPanelItem), (size_t)Size, Item.get() };
      startupInfo.PanelControl(this, FCTL_GETCURRENTPANELITEM, 0, &GetItem);

      const wchar_t *fName = Item->FileName;

      for(; fNum < noFiles; ++fNum)
      {
          if(!lstrcmpW(fName, pcFiles[fNum].name))
          {
              break;
          }
      }

      DWORD oldCheckSum = calculateCheckSum((BYTE*)&files[fNum], sizeof(SCLHdr));
      
      wchar_t name[9] = L"        ";
      std::transform(std::cbegin(files[fNum].name), std::cend(files[fNum].name), name, Cp866ToUcs2);

      wchar_t type[2] = L" ";
      type[0] = files[fNum].type ? files[fNum].type : ' ';

      wchar_t start[6];
      wsprintfW(start, L"%d", files[fNum].start);

      FarDialogItem dialogItems[]=
      {
        { DI_DOUBLEBOX,  3, 1, 34, 7, { FALSE }, nullptr, nullptr,                                   0, GetMsg(MEditFileInfo) },
        { DI_TEXT,       5, 2,  0, 0, { FALSE }, nullptr, nullptr,                                   0, GetMsg(MEditName) },
        { DI_EDIT,      24, 2, 32, 2,  { TRUE }, nullptr, nullptr,                                   0, name, std::size(name) - 1 },
        { DI_TEXT,       5, 3,  0, 0, { FALSE }, nullptr, nullptr,                                   0, GetMsg(MEditType) },
        { DI_EDIT,      24, 3, 25, 3, { FALSE }, nullptr, nullptr,                                   0, type, std::size(type) - 1 },
        { DI_TEXT,       5, 4,  0, 0, { FALSE }, nullptr, nullptr,                                   0, GetMsg(MEditStart) },
        { DI_EDIT,      24, 4, 29, 4, { FALSE }, nullptr, nullptr,                                   0, start, std::size(start) - 1 },
        { DI_TEXT,       3, 5,  0, 0, { FALSE }, nullptr, nullptr,        DIF_BOXCOLOR | DIF_SEPARATOR, L"" },
        { DI_BUTTON,     0, 6,  0, 0, { FALSE }, nullptr, nullptr, DIF_CENTERGROUP | DIF_DEFAULTBUTTON, GetMsg(MOk) },
        { DI_BUTTON,     0, 6,  0, 0, { FALSE }, nullptr, nullptr,                     DIF_CENTERGROUP, GetMsg(MCancel) }
      };

      std::unique_ptr<void, decltype(SafeDlgFree) *> Dlg{ startupInfo.DialogInit(&MainGuid,
                                       &__uuidof(struct __declspec(uuid("{AE4CCA43-335F-40C0-A89D-33DC6B76E4FE}")) TDlgGuid),
                                       -1, -1, 38, 9,
                                       NULL,
                                       dialogItems,
                                       std::size(dialogItems), 0, 0, startupInfo.DefDlgProc, nullptr), SafeDlgFree };
      if(Dlg.get() == INVALID_HANDLE_VALUE)
      {
          return FALSE;
      }
      intptr_t askCode = startupInfo.DialogRun(Dlg.get());

      if(askCode != 8) return TRUE;

      {
          FarDialogItemData ItemData = { sizeof(FarDialogItemData), std::size(name) - 1, name };
          startupInfo.SendDlgMessage(Dlg.get(), DM_GETTEXT, 2, &ItemData);
      }

      std::fill(std::begin(files[fNum].name), std::end(files[fNum].name), ' ');

      {
          unsigned i = 0;
          for(auto c : name)
          {
              if(c == 0) break;
              //   cp866
              WideCharToMultiByte(866, 0, &c, 1, &files[fNum].name[i++], 1, nullptr, nullptr);
          }
      }
      {
          FarDialogItemData ItemData = { sizeof(FarDialogItemData), std::size(type) - 1, type };
          startupInfo.SendDlgMessage(Dlg.get(), DM_GETTEXT, 4, &ItemData);
      }
      files[fNum].type = 'C';
      if(type[0] > 0) files[fNum].type = type[0];

      {
          FarDialogItemData ItemData = { sizeof(FarDialogItemData), std::size(start) - 1, start };
          startupInfo.SendDlgMessage(Dlg.get(), DM_GETTEXT, 6, &ItemData);
      }

      int result = str2int(start, std::size(start) - 1);
      if(result != -1) files[fNum].start = result;
      
      DWORD newCheckSum = calculateCheckSum((BYTE*)&files[fNum], sizeof(SCLHdr));
      
      if(!openHostFile()) return TRUE;
      DWORD noBytesRead, noBytesWritten;

      static_assert(sizeof(files[fNum]) == sizeof(SCLHdr));
      SetFilePointer(hostFile, fNum*sizeof(SCLHdr) + 8 + 1, NULL, FILE_BEGIN);
      WriteFile(hostFile, &files[fNum], sizeof(SCLHdr), &noBytesWritten, NULL);
      SetFilePointer(hostFile, -4, NULL, FILE_END);
      
      DWORD checkSum;
      ReadFile(hostFile, &checkSum, 4, &noBytesRead, NULL);
      checkSum = checkSum - oldCheckSum + newCheckSum;
      SetFilePointer(hostFile, -4, NULL, FILE_END);
      WriteFile(hostFile, &checkSum, 4, &noBytesWritten, NULL);
      
      closeHostFile();
      startupInfo.PanelControl(this, FCTL_UPDATEPANEL, 1, nullptr);
      startupInfo.PanelControl(this, FCTL_REDRAWPANEL, 0, nullptr);
      startupInfo.PanelControl(PANEL_PASSIVE, FCTL_UPDATEPANEL, 1, nullptr);

      return TRUE;
    }
  }
  return FALSE;
}
