#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <string>
#include "plugin.hpp"

#include "types.hpp"
#include "types_trd.hpp"
#include "manager.hpp"
#include "tools.hpp"
#include "lang.hpp"

extern PluginStartupInfo startupInfo;
extern Options           op;
extern const GUID MainGuid;

void addEndSlash(std::wstring &path);


enum Action   { ASK_USER, OVERWRITE_ALL, SKIP_ALL };
static Action userAction;

static bool   skipPathnames;
static bool   skipHeaders;
static bool   scl;
static bool   keepSilence;

static HANDLE file;

  //      SCL
static BYTE   noFilesWritten;

static BYTE*  hdrs;
static BYTE*  bodies;
static DWORD  bodiesSize;

ExitCode createFile(wchar_t* name)
{
  //   
  WIN32_FIND_DATAW data;
  HANDLE h = FindFirstFileW(name, &data);
  if(h != INVALID_HANDLE_VALUE)
  {
    //     
    FindClose(h);
    if(userAction == SKIP_ALL) return SKIP;
    if(userAction == ASK_USER)
    {
      const wchar_t *msgItems[11];
      msgItems[0] = GetMsg(MWarning);
      msgItems[1] = GetMsg(MAlreadyExists);
      msgItems[2] = name;
      msgItems[3] = L"\x001";
      msgItems[5] = L"\x001";
      msgItems[6] = GetMsg(MOverwrite);
      msgItems[7] = GetMsg(MAll);
      msgItems[8] = GetMsg(MSkip);
      msgItems[9] = GetMsg(MSkipAll);
      msgItems[10] = GetMsg(MCancel);
      wchar_t shortName[51];
      msgItems[2] = truncPathStr(shortName, name, 50);
      
      wchar_t param[50];
      if(data.ftLastWriteTime.dwLowDateTime == 0 &&
         data.ftLastWriteTime.dwHighDateTime == 0)
      {
        wsprintfW(param, GetMsg(MDestination1), data.nFileSizeLow);
      }
      else
      {
        FILETIME lastWriteTime;
        FileTimeToLocalFileTime(&data.ftLastWriteTime, &lastWriteTime);
        
        SYSTEMTIME time;
        FileTimeToSystemTime(&lastWriteTime, &time);
        
        wsprintfW(param, GetMsg(MDestination2),
                 data.nFileSizeLow,
                 time.wDay, time.wMonth, time.wYear,
                 time.wHour, time.wMinute, time.wSecond);
      }
      msgItems[4] = param;
      int askCode = messageBox(FMSG_WARNING/* | FMSG_DOWN*/,
                               msgItems,
                               std::size(msgItems), 5);
      switch(askCode)
      {
        case -1:
        case 4:
                  return CANCEL;
        case 3:
                  userAction = SKIP_ALL;
        case 2:
                  return SKIP;
        case 1:
                  userAction = OVERWRITE_ALL;
      }
    }
  }
  file = CreateFileW(name,
                    GENERIC_WRITE,
                    FILE_SHARE_READ,
                    NULL,
                    CREATE_ALWAYS,
                    FILE_FLAG_SEQUENTIAL_SCAN,
                    NULL);
  if(file != INVALID_HANDLE_VALUE) return OK;
  
  const wchar_t * const msgItems[]{ GetMsg(MError), GetMsg(MCanNotCreate), name, GetMsg(MOk) };
  messageBox(FMSG_WARNING, msgItems, std::size(msgItems), 1);
  return SKIP;
}

ExitCode Manager::getFile(int fNum, bool isMove, const wchar_t* to)
{
  ExitCode exitCode = OK;

  wchar_t name[300];
  wsprintfW(name, L"%s%s", to, pcFiles[fNum].name);
  
  wchar_t shortName[50];
  const wchar_t * const msgItems[]{ GetMsg(MCopyingFile), GetMsg(MCopyTheFile), truncPathStr(shortName, name, 30) };
  if(!keepSilence)
    messageBox(0, msgItems, std::size(msgItems), 0);
  
  int lastPartSize = sectorSize;

  if(skipHeaders)
  {
    int size    = files[fNum].size;
    int noSecs2 = size/sectorSize + !!(size%sectorSize);
    if(files[fNum].noSecs == noSecs2 && size%sectorSize) lastPartSize = size % sectorSize;
  }
  
  if(scl)
  {
    hdrs= (BYTE*)realloc(hdrs,   (noFilesWritten+1)*sizeof(SCLHdr));
    memcpy(hdrs+noFilesWritten*sizeof(SCLHdr), &files[fNum], sizeof(SCLHdr));
    if(files[fNum].noSecs)
    {
      bodies = (BYTE*)realloc(bodies, bodiesSize+sectorSize*(files[fNum].noSecs-1)+lastPartSize);

      BYTE  sector[sectorSize];
      BYTE* ptr = bodies + bodiesSize;
      int trk = files[fNum].trk, sec = files[fNum].sec;
      for(int s = 0; s < files[fNum].noSecs-1; ++s)
      {
        read(trk, sec, sector);
        if(++sec == 16) { trk++; sec = 0; }
        memcpy(ptr, sector, sectorSize);
        ptr        += sectorSize;
        bodiesSize += sectorSize;
      }
      //     
      read(trk, sec, sector);
      memcpy(ptr, sector, lastPartSize);
      bodiesSize += lastPartSize;
    }
    ++noFilesWritten;
  }
  else
  {
    if(skipHeaders) lstrcatW(name, L".bin");
    ExitCode code = createFile(name);
    if(code != OK) return code;

    DWORD noBytesWritten;
    if(!skipHeaders)
    {
      HoHdr hdr;
      std::copy_n(files[fNum].name, std::size(hdr.name), hdr.name);

      hdr.type     = files[fNum].type;
      hdr.start    = files[fNum].start;
      hdr.size     = files[fNum].size;
      hdr.reserved = 0;
      hdr.noSecs   = files[fNum].noSecs;
      hdr.checkSum = calculateCheckSum(hdr);
      WriteFile(file, &hdr, sizeof(HoHdr), &noBytesWritten, NULL);
    }

    if(files[fNum].noSecs)
    {
      //   
      BYTE sector[sectorSize];

      int trk = files[fNum].trk, sec = files[fNum].sec;
      for(int s = 0; s < files[fNum].noSecs-1; ++s)
      {
        read(trk, sec, sector);
        if(++sec == 16) { trk++; sec = 0; }
        WriteFile(file, sector, sectorSize, &noBytesWritten, NULL);
      }
      //     
      read(trk, sec, sector);
      WriteFile(file, sector, lastPartSize, &noBytesWritten, NULL);
    }
    CloseHandle(file);
  }
  
  // "" ( )   
  if(isMove && exitCode == OK)
  {
    if(files[fNum].name[0] != 0x01) ++noDelFiles;
    files[fNum].name[0] = 0x01;
  }
  return exitCode;
}

ExitCode Manager::getFolder(int fNum, bool isMove, const wchar_t* to)
{
  ExitCode exitCode = OK;

  wchar_t dest[300];
  lstrcpyW(dest, to);
  if(!skipPathnames)
  {
    lstrcatW(dest, pcFolders[fNum-1]);
    addEndSlash(dest);
    CreateDirectoryW(dest, NULL);
  }
  //   
  for(int i = 0; i < noFolders; ++i)
    if(folderMap[i] == fNum)
    {
      ExitCode code = getFolder(i+1, isMove, dest);
      
      if(code == CANCEL) return CANCEL;
      if(code == SKIP)   exitCode = SKIP;
    }

  //   
  for(int i = 0; i < noFiles; ++i)
    if(fileMap[i] == fNum)
    {
      ExitCode code = getFile(i, isMove, dest);
      if(code == CANCEL) return CANCEL;
      if(code == SKIP)   exitCode = SKIP;
    }

  // "" ( )   
  if(isMove && exitCode == OK)
  {
    if(folders[fNum-1][0] != 0x01) ++noDelFolders;
    folders[fNum-1][0] = 0x01;
  }
  return exitCode;
}

int Manager::getFiles(PluginPanelItem *panelItem,
                      size_t noItems,
                      BOOL isMove,
                      const wchar_t *destPath,
                      OPERATION_MODES opMode)
{
  if(noItems == 0) return 1;
  if(fmt->isProtected(img) && isMove)
  {
    const wchar_t * const msgItems[]{ GetMsg(MWarning), GetMsg(MCanNotMove), GetMsg(MWriteProtected), GetMsg(MOk) };
    messageBox(FMSG_WARNING, msgItems, std::size(msgItems), 1);
    return FALSE;
  }

  std::wstring DestPathCopy(destPath);

  if(!DestPathCopy.empty()) addEndSlash(DestPathCopy);
  
  wchar_t msg[100];
  const wchar_t * const action[]{ GetMsg(MCopy), GetMsg(MMove) };
  const wchar_t * const action2[]{ GetMsg(MCopying), GetMsg(MMoving) };
  const wchar_t * const button[]{ GetMsg(MCopyButton), GetMsg(MMoveButton) };
  if(noItems == 1)
    wsprintfW(msg, GetMsg(MFileTo), action2[!!isMove], panelItem[0].FileName);
  else
    wsprintfW(msg, GetMsg(MFilesTo), action2[!!isMove], noItems);

  const wchar_t *historyName = L"xTRD_copy_path";
  
  FarDialogItem dialogItems[]={
    { DI_DOUBLEBOX,   3, 1, 72, 12, { FALSE }, nullptr,     nullptr,                                   0, action[!!isMove],              0, 0, { 0, 0 } },
    { DI_TEXT,        5, 2,  0,  0, { FALSE }, nullptr,     nullptr,                                   0, msg,                           0, 0, { 0, 0 } },
    { DI_EDIT,        5, 3, 70,  0, { TRUE },  historyName, nullptr,                         DIF_HISTORY, DestPathCopy.data(),           0, 0, { 0, 0 } },
    { DI_TEXT,        3, 4,  0,  0, { FALSE }, nullptr,     nullptr,        DIF_BOXCOLOR | DIF_SEPARATOR, L"",                           0, 0, { 0, 0 } },
    { DI_RADIOBUTTON, 5, 5,  0,  0, { FALSE }, nullptr,     nullptr,                           DIF_GROUP, GetMsg(MHoBeta),               0, 0, { 0, 0 } },
    { DI_RADIOBUTTON, 5, 6,  0,  0, { FALSE }, nullptr,     nullptr,                                   0, GetMsg(MSCL),                  0, 0, { 0, 0 } },
    { DI_TEXT,        3, 7,  0,  0, { FALSE }, nullptr,     nullptr,        DIF_BOXCOLOR | DIF_SEPARATOR, L"",                           0, 0, { 0, 0 } },
    { DI_CHECKBOX,    5, 8,  0,  0, { FALSE }, nullptr,     nullptr,                                   0, GetMsg(MSkipHeaders),          0, 0, { 0, 0 } },
    { DI_CHECKBOX,    5, 9,  0,  0, { FALSE }, nullptr,     nullptr,                                   0, GetMsg(MCopyWithoutPathnames), 0, 0, { 0, 0 } },
    { DI_TEXT,        3, 10, 0,  0, { FALSE }, nullptr,     nullptr,        DIF_BOXCOLOR | DIF_SEPARATOR, L"",                           0, 0, { 0, 0 } },
    { DI_BUTTON,      0, 11, 0,  0, { FALSE }, nullptr,     nullptr, DIF_CENTERGROUP | DIF_DEFAULTBUTTON, button[!!isMove],              0, 0, { 0, 0 } },
    { DI_BUTTON,      0, 11, 0,  0, { FALSE }, nullptr,     nullptr,                     DIF_CENTERGROUP, GetMsg(MCancel),               0, 0, { 0, 0 } }
  };
  
  switch(op.defaultFormat)
  {
    case 1:
              dialogItems[4].Selected = TRUE;
              break;
    case 2:
              dialogItems[5].Selected = TRUE;
              break;
    default:
              if(noItems == 1)
                dialogItems[4].Selected = TRUE;
              else
                dialogItems[5].Selected = TRUE;
  }
  
  scl           = false;
  skipHeaders   = false;
  skipPathnames = false;
  keepSilence   = opMode & OPM_SILENT;

  //    ,  
  if(!keepSilence)
  {
    std::unique_ptr<void, decltype(SafeDlgFree) *> Dlg{ startupInfo.DialogInit(&MainGuid,
                                     &__uuidof(struct __declspec(uuid("{5F17255C-7F56-40ED-B875-A0268EB38433}")) TDlgGuid),
                                     -1, -1, 76, 14,
                                     NULL,
                                     dialogItems,
                                     std::size(dialogItems), 0, 0, startupInfo.DefDlgProc, nullptr), SafeDlgFree };
    if(Dlg.get() == INVALID_HANDLE_VALUE)
    {
        return -1;
    }
    intptr_t askCode = startupInfo.DialogRun(Dlg.get());

    if(askCode != 10) return -1;

    auto Sz = startupInfo.SendDlgMessage(Dlg.get(), DM_GETTEXT, 2, nullptr);
    DestPathCopy.resize(Sz);

    FarDialogItemData ItemData = { sizeof(FarDialogItemData), DestPathCopy.size(), DestPathCopy.data() };
    startupInfo.SendDlgMessage(Dlg.get(), DM_GETTEXT, 2, &ItemData);

    scl           = !startupInfo.SendDlgMessage(Dlg.get(), DM_GETCHECK, 4, 0);
    skipHeaders = startupInfo.SendDlgMessage(Dlg.get(), DM_GETCHECK, 7, 0);
    skipPathnames = startupInfo.SendDlgMessage(Dlg.get(), DM_GETCHECK, 8, 0);
  }
  
  //   ,   
  if(GetFileAttributesW(DestPathCopy.c_str()) == 0xFFFFFFFF)
  {
      for(wchar_t *c = DestPathCopy.data(); *c; c++)
      {
          if(*c != L' ')
          {
              for(; *c; c++)
              {
                  if(*c == L'\\')
                  {
                      *c = 0;
                      CreateDirectoryW(DestPathCopy.data(), NULL);
                      *c = L'\\';
                  }
              }
              CreateDirectoryW(DestPathCopy.data(), NULL);
              break;
          }
      }
  }
  if(!DestPathCopy.empty() && DestPathCopy.back() != ':') addEndSlash(DestPathCopy);

  if(!openHostFile()) return 0;

  userAction = ASK_USER;
  
  if(scl)
  {
    skipPathnames =  true;
    noFilesWritten = 0;
    
    wchar_t name[300];
    wsprintfW(name, L"%s%s", DestPathCopy.c_str(), panelItem[0].FileName);
    
    wchar_t* ptr = pointToName(name);
    while(*ptr && *ptr != L'.') ++ptr;
    *ptr = 0;
    
    if(skipHeaders)
      lstrcatW(name, L".bin");
    else
      lstrcatW(name, L".scl");
    
    ExitCode code = createFile(name);
    
    if(code != OK) return -1;
    
    hdrs           = (BYTE*)malloc(1);
    bodies         = (BYTE*)malloc(1);
    bodiesSize     = 0;
  }
  
  int returnCode = 1;
  
  HANDLE screen = startupInfo.SaveScreen(0,0,-1,-1);
  
  for(int iNum = 0; iNum < noItems; ++iNum)
  {
    ExitCode exitCode;
    int fNum;
    if(panelItem[iNum].FileAttributes & FILE_ATTRIBUTE_DIRECTORY)
    {
      for(fNum = 0; fNum < noFolders; ++fNum)
      {
        if(folderMap[fNum] != curFolderNum) continue;
        if(!lstrcmpW(pcFolders[fNum], panelItem[iNum].FileName)) break;
      }
      exitCode = getFolder(fNum+1, isMove, DestPathCopy.c_str());
    }
    else
    {
      for(fNum = 0; fNum < noFiles; ++fNum)
        if(!lstrcmpW(panelItem[iNum].FileName, pcFiles[fNum].name)) break;

      //     
      if(noItems == 1 && ((opMode & OPM_VIEW) || (opMode & OPM_EDIT)) && pcFiles[fNum].skipHeader) skipHeaders = true;
      exitCode = getFile(fNum, isMove, DestPathCopy.c_str());
    }

    if(exitCode == SKIP)
    {
      returnCode = 0;
      continue;
    }
    if(exitCode == CANCEL)
    {
      returnCode = -1;
      break;
    }
    //  /   
    panelItem[iNum].Flags ^= PPIF_SELECTED;
  }
  startupInfo.RestoreScreen(screen);

  if(scl)
  {
    const wchar_t* const msgItems[]{ GetMsg(MWaiting), GetMsg(MWritingSCL) };
    if(!keepSilence)
    {
      screen = startupInfo.SaveScreen(0,0,-1,-1);
      messageBox(0, msgItems, std::size(msgItems), 0);
      startupInfo.RestoreScreen(screen);
    }
    
    DWORD noBytesWritten;
    DWORD checkSum = 0x255UL + DWORD(noFilesWritten);
    static const BYTE  signature[] = {'S', 'I', 'N', 'C', 'L', 'A', 'I', 'R' };
    if(!skipHeaders)
    {
      WriteFile(file, signature, sizeof(signature), &noBytesWritten, NULL);
      WriteFile(file, &noFilesWritten, 1, &noBytesWritten, NULL);
      checkSum += calculateCheckSum(hdrs, noFilesWritten*sizeof(SCLHdr));
      WriteFile(file, hdrs, noFilesWritten*sizeof(SCLHdr), &noBytesWritten, NULL);
      checkSum += calculateCheckSum(bodies, bodiesSize);
    }
    WriteFile(file, bodies, bodiesSize, &noBytesWritten, NULL);

    if(!skipHeaders)
      WriteFile(file, &checkSum, sizeof(checkSum), &noBytesWritten, NULL);
    
    CloseHandle(file);
    free(bodies);
    free(hdrs);
  }
  if(isMove)
  {
    if(op.autoMove)
      move();
    else
      diskInfo.noDelFiles = noDelFiles;
    
    writeInfo();
  }
  closeHostFile();
  return returnCode;
}
