#define WIN32_LEAN_AND_MEAN
#include <windows.h>

#include "types.hpp"
#include "types_zxzip.hpp"
#include "manager.hpp"
#include "detector.hpp"
#include "tools.hpp"
#include "tools_zxzip.hpp"

extern Detector* detector;
extern Options   op;

bool Manager::openHostFile(void)
{
  hostFile = CreateFileW(hostFileName,
                        GENERIC_READ,
                        FILE_SHARE_READ | FILE_SHARE_WRITE,
                        NULL,
                        OPEN_EXISTING,
                        FILE_FLAG_SEQUENTIAL_SCAN,
                        NULL);
  volumeNum = 0;
  return (hostFile != INVALID_HANDLE_VALUE);
}

void Manager::closeHostFile(void)
{
  if(hostFile != INVALID_HANDLE_VALUE) CloseHandle(hostFile);
}

void Manager::makePCNames(void)
{
  for(int fNum = 0; fNum < noFiles; ++fNum)
  {
    const char* from    = files  [fNum].name;
    wchar_t* to      = pcFiles[fNum].name;

    makePCName(from, to);
    BYTE noChars = lstrlenW(to);


    BYTE dotPos = noChars;
    to[noChars++] = '.';
    to[noChars++] = '$';
    if(isValidChar(files[fNum].type)) to[noChars++] = files[fNum].type;
    to[noChars] = 0;
    
    BYTE typeNum = 0xFF;
    if(op.detectFormat)
    {
      BYTE secs[16*256];
      ZeroMemory(secs, 16*256);

      int noSecs = files[fNum].noSecs;
      if(noSecs > 16) noSecs = 16;
      if(!noSecs)
        if(!extractFile(fNum, secs, noSecs*256)) ZeroMemory(secs, noSecs*256);

      typeNum = detector->detect(files[fNum], secs, noSecs*256, pcFiles[fNum].comment);
    }
    pcFiles[fNum].type = typeNum;

    pcFiles[fNum].skipHeader = false;

    if(typeNum != 0xFF)
    {
      detector->specialChar(typeNum, to+dotPos+1);
      detector->getType(typeNum, to+dotPos+2);
      pcFiles[fNum].skipHeader = detector->getSkipHeader(typeNum);
    }
    //   zxzip 
    if(fNum != 0 &&
       !memcmp(from, "********ZIP", 11) &&
       !memcmp(&files[fNum-1].type, "ZIP", 3))
    {
      if(memcmp(files[fNum-1].name, "********", 8))
        lstrcpyW(to, pcFiles[fNum-1].name);
      else
        lstrcpynW(to, pcFiles[fNum-1].name, lstrlenW(pcFiles[fNum-1].name));
    }
    //   
    int i = fNum;
    while(i-- > 0)
    {
      int len = lstrlenW(to);
      if(len && !wcsnicmp(to, pcFiles[i].name, len))
      {
        wchar_t ch = pcFiles[i].name[len];
        ch = (ch != 0) ? ch+1 : '0';
        
        //     :(
        if(ch == '9'+1) ch = 'A'; 
        if(ch == 'Z'+1) ch = 'a';
        if(ch == 'z'+1) ch = '0';
        
        to[len] = ch; to[len+1] = 0;
        break;
      }
    }
  }
}
